#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest
import unittest.mock as mock
from socket import (
    AF_INET,
    AF_INET6,
)

from neo4j.addressing import (
    Address,
    IPv4Address,
    IPv6Address,
)
from neo4j import GraphDatabase

mock_socket_ipv4 = mock.Mock()
mock_socket_ipv4.getpeername = lambda: ("127.0.0.1", 7687)  # (address, port)

mock_socket_ipv6 = mock.Mock()
mock_socket_ipv6.getpeername = lambda: ("[::1]", 7687, 0, 0)  # (address, port, flow info, scope id)

# python -m pytest tests/unit/test_addressing.py -s


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("127.0.0.1", 7687), {"family": AF_INET, "host": "127.0.0.1", "port": 7687, "str": "127.0.0.1:7687", "repr": "IPv4Address(('127.0.0.1', 7687))"}),
        (("localhost", 7687), {"family": AF_INET, "host": "localhost", "port": 7687, "str": "localhost:7687", "repr": "IPv4Address(('localhost', 7687))"}),
        ((None, None), {"family": AF_INET, "host": None, "port": None, "str": "None:None", "repr": "IPv4Address((None, None))"}),
        (("::1", 7687), {"family": AF_INET, "host": "::1", "port": 7687, "str": "::1:7687", "repr": "IPv4Address(('::1', 7687))"}),
        (("::1", 7687, 0, 0), {"family": AF_INET6, "host": "::1", "port": 7687, "str": "[::1]:7687", "repr": "IPv6Address(('::1', 7687, 0, 0))"}),
        (("::1", 7687, 1, 2), {"family": AF_INET6, "host": "::1", "port": 7687, "str": "[::1]:7687", "repr": "IPv6Address(('::1', 7687, 1, 2))"}),
        ((None, None, None, None), {"family": AF_INET6, "host": None, "port": None, "str": "[None]:None", "repr": "IPv6Address((None, None, None, None))"}),
        (Address(("127.0.0.1", 7687)), {"family": AF_INET, "host": "127.0.0.1", "port": 7687, "str": "127.0.0.1:7687", "repr": "IPv4Address(('127.0.0.1', 7687))"}),
        (Address(("::1", 7687, 1, 2)), {"family": AF_INET6, "host": "::1", "port": 7687, "str": "[::1]:7687", "repr": "IPv6Address(('::1', 7687, 1, 2))"}),
    ]
)
def test_address_initialization(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_initialization
    address = Address(test_input)
    assert address.family == expected["family"]
    assert address.host == expected["host"]
    assert address.port == expected["port"]
    assert str(address) == expected["str"]
    assert repr(address) == expected["repr"]


@pytest.mark.parametrize(
    "test_input",
    [
        Address(("127.0.0.1", 7687)),
        Address(("127.0.0.1", 7687, 1, 2)),
    ]
)
def test_address_init_with_address_object_returns_same_instance(test_input):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_init_with_address_object_returns_same_instance
    address = Address(test_input)
    assert address is test_input
    assert id(address) == id(test_input)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("127.0.0.1",), ValueError),
        (("127.0.0.1", 7687, 0), ValueError),
        (("[::1]", 7687, 0), ValueError),
        (("[::1]", 7687, 0, 0, 0), ValueError),
    ]
)
def test_address_initialization_with_incorrect_input(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_initialization_with_incorrect_input
    with pytest.raises(expected):
        address = Address(test_input)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (mock_socket_ipv4, ("127.0.0.1", 7687)),
        (mock_socket_ipv6, ("[::1]", 7687, 0, 0))
    ]
)
def test_address_from_socket(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_from_socket

    address = Address.from_socket(test_input)
    assert address == expected


def test_address_from_socket_with_none():
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_from_socket_with_none
    with pytest.raises(AttributeError):
        address = Address.from_socket(None)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ("127.0.0.1:7687", ("127.0.0.1", 7687)),
        ("localhost:7687", ("localhost", 7687)),
        (":7687", ("localhost", 7687)),
        (":", ("localhost", 0)),
        ("", ("localhost", 0)),
        (":abcd", ("localhost", "abcd")),
        (" ", (" ", 0)),
    ]
)
def test_address_parse_with_ipv4(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_parse_with_ipv4
    parsed = Address.parse(test_input)
    assert parsed == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        ("[::1]:7687", ("::1", 7687, 0, 0)),
        ("[::1]:abcd", ("::1", "abcd", 0, 0)),
        ("[::1]:", ("::1", 0, 0, 0)),
        ("[::1]", ("::1", 0, 0, 0)),
    ]
)
def test_address_should_parse_ipv6(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_should_parse_ipv6
    parsed = Address.parse(test_input)
    assert parsed == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (None, TypeError),
        (123, TypeError),
        (("127.0.0.1", 7687), TypeError),
        (("[::1]", 7687, 1, 2), TypeError),
        (Address(("127.0.0.1", 7687)), TypeError),
    ]
)
def test_address_parse_with_invalid_input(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_parse_with_invalid_input
    with pytest.raises(expected):
        parsed = Address.parse(test_input)


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("localhost:7687 [::1]:7687",), 2),
        (("localhost:7687", "[::1]:7687"), 2),
        (("localhost:7687 localhost:7688", "[::1]:7687"), 3),
        (("localhost:7687 localhost:7687", "[::1]:7687"), 3),
    ]
)
def test_address_parse_list(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_parse_list
    addresses = Address.parse_list(*test_input)
    assert len(addresses) == expected


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (("localhost:7687", None), TypeError),
        (("localhost:7687", 123), TypeError),
        (("localhost:7687", ("127.0.0.1", 7687)), TypeError),
        (("localhost:7687", ("[::1]", 7687, 1, 2)), TypeError),
        (("localhost:7687", Address(("127.0.0.1", 7687))), TypeError),
    ]
)
def test_address_parse_list_with_invalid_input(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_parse_list_with_invalid_input
    with pytest.raises(TypeError):
        addresses = Address.parse_list(*test_input)


def test_address_resolve():
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_resolve
    address = Address(("127.0.0.1", 7687))
    resolved = address.resolve()
    assert isinstance(resolved, Address) is False
    assert isinstance(resolved, list) is True
    assert len(resolved) == 1
    assert resolved[0] == IPv4Address(('127.0.0.1', 7687))


def test_address_resolve_with_custom_resolver_none():
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_resolve_with_custom_resolver_none
    address = Address(("127.0.0.1", 7687))
    resolved = address.resolve(resolver=None)
    assert isinstance(resolved, Address) is False
    assert isinstance(resolved, list) is True
    assert len(resolved) == 1
    assert resolved[0] == IPv4Address(('127.0.0.1', 7687))


@pytest.mark.parametrize(
    "test_input, expected",
    [
        (Address(("127.0.0.1", "abcd")), ValueError),
        (Address((None, None)), ValueError),
    ]
)
def test_address_resolve_with_unresolvable_address(test_input, expected):
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_resolve_with_unresolvable_address
    with pytest.raises(expected):
        test_input.resolve(resolver=None)


def test_address_resolve_with_custom_resolver():
    # python -m pytest tests/unit/test_addressing.py -s -k test_address_resolve_with_custom_resolver
    custom_resolver = lambda a: [("127.0.0.1", 7687), ("localhost", 1234)]

    address = Address(("127.0.0.1", 7687))
    resolved = address.resolve(family=AF_INET, resolver=custom_resolver)
    assert isinstance(resolved, Address) is False
    assert isinstance(resolved, list) is True
    if len(resolved) == 2:
        # IPv4 only
        assert resolved[0] == IPv4Address(('127.0.0.1', 7687))
        assert resolved[1] == IPv4Address(('127.0.0.1', 1234))
    else:
        assert False



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2018 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import deque
from threading import Thread, Lock
from time import sleep

from neo4j.work import Workspace
from neo4j.conf import WorkspaceConfig
from neo4j.api import (
    WRITE_ACCESS,
)

class PipelineConfig(WorkspaceConfig):

    #:
    flush_every = 8192  # bytes


class Pipeline(Workspace):

    def __init__(self, pool, config):
        assert isinstance(config, PipelineConfig)
        super(Pipeline, self).__init__(pool, config)
        self._connect(WRITE_ACCESS)
        self._flush_every = config.flush_every
        self._data = deque()
        self._pull_lock = Lock()

    def push(self, statement, parameters=None):
        self._connection.run(statement, parameters)
        self._connection.pull(on_records=self._data.extend)
        output_buffer_size = len(self._connection.outbox.view())
        if output_buffer_size >= self._flush_every:
            self._connection.send_all()

    def _results_generator(self):
        results_returned_count = 0
        try:
            summary = 0
            while summary == 0:
                _, summary = self._connection.fetch_message()
            summary = 0
            while summary == 0:
                detail, summary = self._connection.fetch_message()
                for n in range(detail):
                    response = self._data.popleft()
                    results_returned_count += 1
                    yield response
        finally:
            self._pull_lock.release()

    def pull(self):
        """Returns a generator containing the results of the next query in the pipeline"""
        # n.b. pull is now somewhat misleadingly named because it doesn't do anything
        # the connection isn't touched until you try and iterate the generator we return
        lock_acquired = self._pull_lock.acquire(blocking=False)
        if not lock_acquired:
            raise PullOrderException()
        return self._results_generator()


class PullOrderException(Exception):
    """Raise when calling pull if a previous pull result has not been fully consumed"""


class Pusher(Thread):

    def __init__(self, pipeline):
        super(Pusher, self).__init__()
        self.pipeline = pipeline
        self.running = True
        self.count = 0

    def run(self):
        while self.running:
            self.pipeline.push("RETURN $x", {"x": self.count})
            self.count += 1


class Puller(Thread):

    def __init__(self, pipeline):
        super(Puller, self).__init__()
        self.pipeline = pipeline
        self.running = True
        self.count = 0

    def run(self):
        while self.running:
            for _ in self.pipeline.pull():
                pass    # consume and discard records
            self.count += 1


def main():
    from neo4j import Driver
    # from neo4j.bolt.diagnostics import watch
    # watch("neobolt")
    with Driver("bolt://", auth=("neo4j", "password")) as dx:
        p = dx.pipeline(flush_every=1024)
        pusher = Pusher(p)
        puller = Puller(p)
        try:
            pusher.start()
            puller.start()
            while True:
                print("sent %d, received %d, backlog %d" % (pusher.count, puller.count, pusher.count - puller.count))
                sleep(1)
        except KeyboardInterrupt:
            pusher.running = False
            pusher.join()
            puller.running = False
            puller.join()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from contextlib import redirect_stdout
from io import StringIO


from neo4j import GraphDatabase
# tag::database-selection-import[]
from neo4j import READ_ACCESS
# end::database-selection-import[]

from neo4j.exceptions import ServiceUnavailable
from neo4j._exceptions import BoltHandshakeError


# python -m pytest tests/integration/examples/test_database_selection_example.py -s -v


class DatabaseSelectionExample:

    def __init__(self, uri, user, password):
        self.driver = GraphDatabase.driver(uri, auth=(user, password))
        with self.driver.session(database="system") as session:
            session.run("DROP DATABASE example IF EXISTS").consume()
            session.run("CREATE DATABASE example").consume()

    def close(self):
        with self.driver.session(database="system") as session:
            session.run("DROP DATABASE example").consume()
        self.driver.close()

    def run_example_code(self):

        driver = self.driver
        # tag::database-selection[]
        with driver.session(database="example") as session:
            session.run("CREATE (a:Greeting {message: 'Hello, Example-Database'}) RETURN a").consume()

        with driver.session(database="example", default_access_mode=READ_ACCESS) as session:
            message = session.run("MATCH (a:Greeting) RETURN a.message as msg").single().get("msg")
            print(message)
        # end::database-selection[]


def test_database_selection_example(neo4j_uri, auth):
    try:
        s = StringIO()
        with redirect_stdout(s):
            example = DatabaseSelectionExample(neo4j_uri, auth[0], auth[1])
            example.run_example_code()
            example.close()
        assert s.getvalue().startswith("Hello, Example-Database")
    except ServiceUnavailable as error:
        if isinstance(error.__cause__, BoltHandshakeError):
            pytest.skip(error.args[0])
        if error.args[0] == "Server does not support routing":
            # This is because a single instance Neo4j 3.5 does not have dbms.routing.cluster.getRoutingTable() call
            pytest.skip(error.args[0])
    except ConfigurationError as error:
            pytest.skip(error.args[0])




#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j.exceptions import (
    ServiceUnavailable,
    ConfigurationError,
    DriverError,
)
from neo4j._exceptions import (
    BoltHandshakeError,
    BoltSecurityError,
)

from neo4j import (
    GraphDatabase,
    BoltDriver,
    Query,
    WRITE_ACCESS,
    READ_ACCESS,
    TRUST_ALL_CERTIFICATES,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    DEFAULT_DATABASE,
    Result,
    unit_of_work,
    Transaction,
)

from tests.stub.conftest import (
    StubCluster,
)

# python -m pytest tests/stub/test_directdriver.py -s -v


driver_config = {
    "encrypted": False,
    "user_agent": "test",
    "max_connection_lifetime": 1000,
    "max_connection_pool_size": 10,
    "keep_alive": True,
    "resolver": None,
}


session_config = {
    "default_access_mode": READ_ACCESS,
    "connection_acquisition_timeout": 1.0,
    "max_transaction_retry_time": 1.0,
    "initial_retry_delay": 1.0,
    "retry_delay_multiplier": 1.0,
    "retry_delay_jitter_factor": 0.1,
    "fetch_size": -1,
}


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/empty.script",
        "v4x0/empty.script",
    ]
)
def test_bolt_uri_constructs_bolt_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_uri_constructs_bolt_driver
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            assert isinstance(driver, BoltDriver)


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/empty_explicit_hello_goodbye.script",
        "v4x0/empty_explicit_hello_goodbye.script",
        "v4x1/empty_explicit_hello_goodbye.script",
        "v4x2/empty_explicit_hello_goodbye.script",
    ]
)
def test_direct_driver_handshake_negotiation(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_driver_handshake_negotiation
    with StubCluster(test_script):
        uri = "bolt://localhost:9001"
        driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config)
        assert isinstance(driver, BoltDriver)
        driver.close()


@pytest.mark.parametrize(
    "test_script, test_expected",
    [
        ("v3/return_1_port_9001.script", "Neo4j/3.0.0"),
        ("v4x0/return_1_port_9001.script", "Neo4j/4.0.0"),
        ("v4x1/return_1_port_9001_bogus_server.script", DriverError),
        ("v4x2/return_1_port_9001_bogus_server.script", DriverError),
    ]
)
def test_return_1_as_x(driver_info, test_script, test_expected):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_return_1_as_x
    with StubCluster(test_script):
        uri = "bolt://localhost:9001"
        try:
            driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test")
            assert isinstance(driver, BoltDriver)
            with driver.session(default_access_mode=READ_ACCESS, fetch_size=-1) as session:
                result = session.run("RETURN 1 AS x")
                value = result.single().value()
                assert value == 1
                summary = result.consume()
                assert summary.server.agent == test_expected
                assert summary.server.agent.startswith("Neo4j")
            driver.close()
        except DriverError as error:
            assert isinstance(error, test_expected)


def test_direct_driver_with_wrong_port(driver_info):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_driver_with_wrong_port
    uri = "bolt://127.0.0.1:9002"
    with pytest.raises(ServiceUnavailable):
        driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config)


@pytest.mark.parametrize(
    "test_script, test_expected",
    [
        ("v3/return_1_port_9001.script", "Neo4j/3.0.0"),
        ("v4x0/return_1_port_9001.script", "Neo4j/4.0.0"),
    ]
)
def test_direct_verify_connectivity(driver_info, test_script, test_expected):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_verify_connectivity
    with StubCluster(test_script):
        uri = "bolt://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            assert isinstance(driver, BoltDriver)
            assert driver.verify_connectivity(default_access_mode=READ_ACCESS) == test_expected


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/disconnect_on_run.script",
        "v4x0/disconnect_on_run.script",
    ]
)
def test_direct_verify_connectivity_disconnect_on_run(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_verify_connectivity_disconnect_on_run
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with pytest.raises(ServiceUnavailable):
                driver.verify_connectivity(default_access_mode=READ_ACCESS)


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/disconnect_on_run.script",
        "v4x0/disconnect_on_run.script",
    ]
)
def test_direct_disconnect_on_run(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_disconnect_on_run
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with pytest.raises(ServiceUnavailable):
                with driver.session(**session_config) as session:
                    session.run("RETURN 1 AS x").consume()


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/disconnect_on_pull_all.script",
        "v4x0/disconnect_on_pull.script",
    ]
)
def test_direct_disconnect_on_pull_all(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_disconnect_on_pull_all
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with pytest.raises(ServiceUnavailable):
                with driver.session(**session_config) as session:
                    session.run("RETURN $x", {"x": 1}).consume()


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/disconnect_after_init.script",
        "v4x0/disconnect_after_init.script",
    ]
)
def test_direct_session_close_after_server_close(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_session_close_after_server_close
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"

        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(**session_config) as session:
                with pytest.raises(ServiceUnavailable):
                    session.write_transaction(lambda tx: tx.run(Query("CREATE (a:Item)", timeout=1)))


@pytest.mark.skip(reason="Cant close the Stub Server gracefully")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/empty.script",
        "v4x0/empty.script",
    ]
)
def test_bolt_uri_scheme_self_signed_certificate_constructs_bolt_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_uri_scheme_self_signed_certificate_constructs_bolt_driver

    test_config = {
        "user_agent": "test",
        "max_connection_lifetime": 1000,
        "max_connection_pool_size": 10,
        "keep_alive": False,
        "max_transaction_retry_time": 1,
        "resolver": None,
    }

    with StubCluster(test_script):
        uri = "bolt+ssc://127.0.0.1:9001"
        try:
            driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **test_config)
            assert isinstance(driver, BoltDriver)
            driver.close()
        except ServiceUnavailable as error:
            assert isinstance(error.__cause__, BoltSecurityError)
            pytest.skip("Failed to establish encrypted connection")


@pytest.mark.skip(reason="Cant close the Stub Server gracefully")
@pytest.mark.parametrize(
    "test_script",
    [
        "v3/empty.script",
        "v4x0/empty.script",
    ]
)
def test_bolt_uri_scheme_secure_constructs_bolt_driver(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_uri_scheme_secure_constructs_bolt_driver

    test_config = {
        "user_agent": "test",
        "max_connection_lifetime": 1000,
        "max_connection_pool_size": 10,
        "keep_alive": False,
        "max_transaction_retry_time": 1,
        "resolver": None,
    }

    with StubCluster(test_script):
        uri = "bolt+s://127.0.0.1:9001"
        try:
            driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **test_config)
            assert isinstance(driver, BoltDriver)
            driver.close()
        except ServiceUnavailable as error:
            assert isinstance(error.__cause__, BoltSecurityError)
            pytest.skip("Failed to establish encrypted connection")


@pytest.mark.parametrize(
    "test_uri",
    [
        "bolt+ssc://127.0.0.1:9001",
        "bolt+s://127.0.0.1:9001",
    ]
)
@pytest.mark.parametrize(
    "test_config, expected_failure, expected_failure_message",
    [
        ({"encrypted": False}, ConfigurationError, "The config settings"),
        ({"encrypted": True}, ConfigurationError, "The config settings"),
        ({"encrypted": True, "trust": TRUST_ALL_CERTIFICATES}, ConfigurationError, "The config settings"),
        ({"trust": TRUST_ALL_CERTIFICATES}, ConfigurationError, "The config settings"),
        ({"trust": TRUST_SYSTEM_CA_SIGNED_CERTIFICATES}, ConfigurationError, "The config settings"),
    ]
)
def test_bolt_uri_scheme_secure_constructs_driver_config_error(driver_info, test_uri, test_config, expected_failure, expected_failure_message):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_uri_scheme_secure_constructs_driver_config_error
    with pytest.raises(expected_failure) as error:
        driver = GraphDatabase.driver(test_uri, auth=driver_info["auth_token"], **test_config)

    assert error.match(expected_failure_message)


@pytest.mark.parametrize(
    "test_config, expected_failure, expected_failure_message",
    [
        ({"trust": 1}, ConfigurationError, "The config setting `trust`"),
        ({"trust": True}, ConfigurationError, "The config setting `trust`"),
        ({"trust": None}, ConfigurationError, "The config setting `trust`"),
    ]
)
def test_driver_trust_config_error(driver_info, test_config, expected_failure, expected_failure_message):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_driver_trust_config_error
    uri = "bolt://127.0.0.1:9001"
    with pytest.raises(expected_failure) as error:
        driver = GraphDatabase.driver(uri, auth=driver_info["auth_token"], **test_config)

    assert error.match(expected_failure_message)


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_n_port_9001.script", "test"),
        ("v4x0/pull_n_port_9001_slow_network.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_normal_case(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_normal_case
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                assert isinstance(result, Result)
                for record in result:
                    expected.append(record["x"])

        assert expected == [1, 2, 3, 4]


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_n_port_9001.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_result_implements_iterator_protocol(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_result_implements_iterator_protocol
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result = iter(session.run("UNWIND [1,2,3,4] AS x RETURN x"))
                expected.append(next(result)["x"])
                expected.append(next(result)["x"])
                expected.append(next(result)["x"])
                expected.append(next(result)["x"])
                with pytest.raises(StopIteration):
                    expected.append(next(result)["x"])

        assert expected == [1, 2, 3, 4]


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_consume_case(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_consume_case
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                result_summary = result.consume()

    assert result_summary.server.address == ('127.0.0.1', 9001)


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_on_exit_consume_remaining_result(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_on_exit_consume_remaining_result
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_on_close_result_consume(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_on_close_result_consume
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            session = driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS)
            result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
            session.close()


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_on_exit_result_consume(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_on_exit_result_consume
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001.script", DEFAULT_DATABASE),
        ("v4x0/pull_n_port_9001.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_case_result_consume(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_case_result_consume
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                for record in result:
                    expected.append(record["x"])

                result_summary = result.consume()
                assert result_summary.server.address == ('127.0.0.1', 9001)

                result_summary = result.consume()  # It will just return the result summary

        assert expected == [1, 2, 3, 4]
        assert result_summary.server.address == ('127.0.0.1', 9001)


@pytest.mark.skip(reason="Bolt Stub Server cant handle this script correctly, see tests/integration/test_bolt_driver.py test_bolt_driver_fetch_size_config_run_consume_run")
@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v4x0/pull_2_discard_all_pull_n_port_9001.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_run_consume_run(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_run_consume_run
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result1 = session.run("UNWIND [1,2,3,4] AS x RETURN x")
                result1.consume()
                result2 = session.run("UNWIND [5,6,7,8] AS x RETURN x")

                for record in result2:
                    expected.append(record["x"])

                result_summary = result2.consume()
                assert result_summary.server.address == ('127.0.0.1', 9001)

        assert expected == [5, 6, 7, 8]
        assert result_summary.server.address == ('127.0.0.1', 9001)


@pytest.mark.skip(reason="Bolt Stub Server cant handle this script correctly, see tests/integration/test_bolt_driver.py test_bolt_driver_fetch_size_config_run_run")
@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v4x0/pull_2_discard_all_pull_n_port_9001.script", "test"),
    ]
)
def test_bolt_driver_fetch_size_config_run_run(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_fetch_size_config_run_run
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = []
                result1 = session.run("UNWIND [1,2,3,4] AS x RETURN x")  # The session should discard all if in streaming mode and a new run is invoked.
                result2 = session.run("UNWIND [5,6,7,8] AS x RETURN x")

                for record in result2:
                    expected.append(record["x"])

                result_summary = result2.consume()
                assert result_summary.server.address == ('127.0.0.1', 9001)

        assert expected == [5, 6, 7, 8]
        assert result_summary.server.address == ('127.0.0.1', 9001)


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v3/pull_all_port_9001_transaction_function.script", DEFAULT_DATABASE),
        ("v4x0/tx_pull_n_port_9001.script", "test"),
        ("v4x0/tx_pull_n_port_9001_slow_network.script", "test"),
        # TODO: Test retry mechanism
    ]
)
def test_bolt_driver_read_transaction_fetch_size_config_normal_case(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_read_transaction_fetch_size_config_normal_case
    @unit_of_work(timeout=3, metadata={"foo": "bar"})
    def unwind(transaction):
        assert isinstance(transaction, Transaction)
        values = []
        result = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
        assert isinstance(result, Result)
        for record in result:
            values.append(record["x"])
        return values

    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                expected = session.read_transaction(unwind)

        assert expected == [1, 2, 3, 4]


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v4x0/tx_pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_explicit_transaction_consume_result_case_a(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_explicit_transaction_consume_result_case_a
    # This test is to check that the implicit consume that is triggered on transaction.commit() is working properly.
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                transaction = session.begin_transaction(timeout=3, metadata={"foo": "bar"})
                result = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
                transaction.commit()


@pytest.mark.parametrize(
    "test_script, database",
    [
        ("v4x0/tx_pull_2_discard_all_port_9001.script", "test"),
    ]
)
def test_bolt_driver_explicit_transaction_consume_result_case_b(driver_info, test_script, database):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_bolt_driver_explicit_transaction_consume_result_case_b
    # This test is to check that the implicit consume that is triggered on transaction.commit() is working properly.
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], user_agent="test") as driver:
            with driver.session(database=database, fetch_size=2, default_access_mode=READ_ACCESS) as session:
                transaction = session.begin_transaction(timeout=3, metadata={"foo": "bar"})
                result = transaction.run("UNWIND [1,2,3,4] AS x RETURN x")
                result.consume()
                transaction.commit()


@pytest.mark.parametrize(
    "test_script",
    [
        "v4x1/return_1_noop_port_9001.script",
        "v4x2/return_1_noop_port_9001.script",
    ]
)
def test_direct_can_handle_noop(driver_info, test_script):
    # python -m pytest tests/stub/test_directdriver.py -s -v -k test_direct_can_handle_noop
    with StubCluster(test_script):
        uri = "bolt://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            assert isinstance(driver, BoltDriver)
            with driver.session(fetch_size=2, default_access_mode=READ_ACCESS) as session:
                result = session.run("RETURN 1 AS x")
                value = result.single().value()
                assert value == 1



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import (
    GraphDatabase,
    WRITE_ACCESS,
    TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
)
from neo4j.exceptions import ServiceUnavailable

from tests.stub.conftest import StubCluster

# python -m pytest tests/stub/test_transactions.py -s -v


driver_config = {
    "encrypted": False,
    "trust": TRUST_SYSTEM_CA_SIGNED_CERTIFICATES,
    "user_agent": "test",
    "max_connection_lifetime": 1000,
    "max_connection_pool_size": 10,
    "keep_alive": False,
    "max_transaction_retry_time": 1,
    "resolver": None,
}

session_config = {
    # "fetch_size": 100,
    # "database": "default",
    # "bookmarks": ["bookmark-1", ],
    "default_access_mode": WRITE_ACCESS,
    "connection_acquisition_timeout": 1.0,
    "max_transaction_retry_time": 1.0,
    "initial_retry_delay": 1.0,
    "retry_delay_multiplier": 1.0,
    "retry_delay_jitter_factor": 0.1,
    "fetch_size": -1,
}





@pytest.mark.parametrize(
    "test_script",
    [
        "v3/connection_error_on_commit.script",
        "v4x0/tx_connection_error_on_commit.script",
    ]
)
def test_connection_error_on_explicit_commit(driver_info, test_script):
    # python -m pytest tests/stub/test_transactions.py -s -v -k test_connection_error_on_explicit_commit
    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(**session_config) as session:
                tx = session.begin_transaction()
                result = tx.run("CREATE (n {name:'Bob'})")
                _ = list(result)
                with pytest.raises(ServiceUnavailable):
                    tx.commit()


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/connection_error_on_commit.script",
        "v4x0/tx_connection_error_on_commit.script",
    ]
)
def test_connection_error_on_commit(driver_info, test_script):
    # python -m pytest tests/stub/test_transactions.py -s -v -k test_connection_error_on_commit
    def create_bob(tx):
        result = tx.run("CREATE (n {name:'Bob'})")
        return list(result)

    with StubCluster(test_script):
        uri = "bolt://127.0.0.1:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"], **driver_config) as driver:
            with driver.session(**session_config) as session:
                with pytest.raises(ServiceUnavailable):
                    session.write_transaction(create_bob)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import TestCase

from neo4j.io import (
    Bolt,
    Neo4jPool,
)
from neo4j.routing import (
    OrderedSet,
    RoutingTable,
)
from neo4j.api import (
    DEFAULT_DATABASE,
)

VALID_ROUTING_RECORD = {
    "ttl": 300,
    "servers": [
        {"role": "ROUTE", "addresses": ["127.0.0.1:9001", "127.0.0.1:9002", "127.0.0.1:9003"]},
        {"role": "READ", "addresses": ["127.0.0.1:9004", "127.0.0.1:9005"]},
        {"role": "WRITE", "addresses": ["127.0.0.1:9006"]},
    ],
}

VALID_ROUTING_RECORD_WITH_EXTRA_ROLE = {
    "ttl": 300,
    "servers": [
        {"role": "ROUTE", "addresses": ["127.0.0.1:9001", "127.0.0.1:9002", "127.0.0.1:9003"]},
        {"role": "READ", "addresses": ["127.0.0.1:9004", "127.0.0.1:9005"]},
        {"role": "WRITE", "addresses": ["127.0.0.1:9006"]},
        {"role": "MAGIC", "addresses": ["127.0.0.1:9007"]},
    ],
}


class OrderedSetTestCase(TestCase):
    def test_should_repr_as_set(self):
        s = OrderedSet([1, 2, 3])
        assert repr(s) == "{1, 2, 3}"

    def test_should_contain_element(self):
        s = OrderedSet([1, 2, 3])
        assert 2 in s

    def test_should_not_contain_non_element(self):
        s = OrderedSet([1, 2, 3])
        assert 4 not in s

    def test_should_be_able_to_get_item_if_empty(self):
        s = OrderedSet([])
        with self.assertRaises(IndexError):
            _ = s[0]

    def test_should_be_able_to_get_items_by_index(self):
        s = OrderedSet([1, 2, 3])
        self.assertEqual(s[0], 1)
        self.assertEqual(s[1], 2)
        self.assertEqual(s[2], 3)

    def test_should_be_iterable(self):
        s = OrderedSet([1, 2, 3])
        assert list(iter(s)) == [1, 2, 3]

    def test_should_have_length(self):
        s = OrderedSet([1, 2, 3])
        assert len(s) == 3

    def test_should_be_able_to_add_new(self):
        s = OrderedSet([1, 2, 3])
        s.add(4)
        assert list(s) == [1, 2, 3, 4]

    def test_should_be_able_to_add_existing(self):
        s = OrderedSet([1, 2, 3])
        s.add(2)
        assert list(s) == [1, 2, 3]

    def test_should_be_able_to_clear(self):
        s = OrderedSet([1, 2, 3])
        s.clear()
        assert list(s) == []

    def test_should_be_able_to_discard_existing(self):
        s = OrderedSet([1, 2, 3])
        s.discard(2)
        assert list(s) == [1, 3]

    def test_should_be_able_to_discard_non_existing(self):
        s = OrderedSet([1, 2, 3])
        s.discard(4)
        assert list(s) == [1, 2, 3]

    def test_should_be_able_to_remove_existing(self):
        s = OrderedSet([1, 2, 3])
        s.remove(2)
        assert list(s) == [1, 3]

    def test_should_not_be_able_to_remove_non_existing(self):
        s = OrderedSet([1, 2, 3])
        with self.assertRaises(ValueError):
            s.remove(4)

    def test_should_be_able_to_update(self):
        s = OrderedSet([1, 2, 3])
        s.update([3, 4, 5])
        assert list(s) == [1, 2, 3, 4, 5]

    def test_should_be_able_to_replace(self):
        s = OrderedSet([1, 2, 3])
        s.replace([3, 4, 5])
        assert list(s) == [3, 4, 5]


class RoutingTableConstructionTestCase(TestCase):
    def test_should_be_initially_stale(self):
        table = RoutingTable(database=DEFAULT_DATABASE)
        assert not table.is_fresh(readonly=True)
        assert not table.is_fresh(readonly=False)


class RoutingTableParseRoutingInfoTestCase(TestCase):
    def test_should_return_routing_table_on_valid_record(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD["servers"],
            ttl=VALID_ROUTING_RECORD["ttl"],
        )
        assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
        assert table.readers == {('127.0.0.1', 9004), ('127.0.0.1', 9005)}
        assert table.writers == {('127.0.0.1', 9006)}
        assert table.ttl == 300

    def test_should_return_routing_table_on_valid_record_with_extra_role(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD_WITH_EXTRA_ROLE["servers"],
            ttl=VALID_ROUTING_RECORD_WITH_EXTRA_ROLE["ttl"],
        )
        assert table.routers == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003)}
        assert table.readers == {('127.0.0.1', 9004), ('127.0.0.1', 9005)}
        assert table.writers == {('127.0.0.1', 9006)}
        assert table.ttl == 300


class RoutingTableServersTestCase(TestCase):
    def test_should_return_all_distinct_servers_in_routing_table(self):
        routing_table = {
            "ttl": 300,
            "servers": [
                {"role": "ROUTE", "addresses": ["127.0.0.1:9001", "127.0.0.1:9002", "127.0.0.1:9003"]},
                {"role": "READ", "addresses": ["127.0.0.1:9001", "127.0.0.1:9005"]},
                {"role": "WRITE", "addresses": ["127.0.0.1:9002"]},
            ],
        }
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=routing_table["servers"],
            ttl=routing_table["ttl"],
        )
        assert table.servers() == {('127.0.0.1', 9001), ('127.0.0.1', 9002), ('127.0.0.1', 9003), ('127.0.0.1', 9005)}


class RoutingTableFreshnessTestCase(TestCase):
    def test_should_be_fresh_after_update(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD["servers"],
            ttl=VALID_ROUTING_RECORD["ttl"],
        )
        assert table.is_fresh(readonly=True)
        assert table.is_fresh(readonly=False)

    def test_should_become_stale_on_expiry(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD["servers"],
            ttl=VALID_ROUTING_RECORD["ttl"],
        )
        table.ttl = 0
        assert not table.is_fresh(readonly=True)
        assert not table.is_fresh(readonly=False)

    def test_should_become_stale_if_no_readers(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD["servers"],
            ttl=VALID_ROUTING_RECORD["ttl"],
        )
        table.readers.clear()
        assert not table.is_fresh(readonly=True)
        assert table.is_fresh(readonly=False)

    def test_should_become_stale_if_no_writers(self):
        table = RoutingTable.parse_routing_info(
            database=DEFAULT_DATABASE,
            servers=VALID_ROUTING_RECORD["servers"],
            ttl=VALID_ROUTING_RECORD["ttl"],
        )
        table.writers.clear()
        assert table.is_fresh(readonly=True)
        assert not table.is_fresh(readonly=False)


class RoutingTableUpdateTestCase(TestCase):
    def setUp(self):
        self.table = RoutingTable(
            database=DEFAULT_DATABASE,
            routers=[("192.168.1.1", 7687), ("192.168.1.2", 7687)],
            readers=[("192.168.1.3", 7687)],
            writers=[],
            ttl=0,
        )
        self.new_table = RoutingTable(
            database=DEFAULT_DATABASE,
            routers=[("127.0.0.1", 9001), ("127.0.0.1", 9002), ("127.0.0.1", 9003)],
            readers=[("127.0.0.1", 9004), ("127.0.0.1", 9005)],
            writers=[("127.0.0.1", 9006)],
            ttl=300,
        )

    def test_update_should_replace_routers(self):
        self.table.update(self.new_table)
        assert self.table.routers == {("127.0.0.1", 9001), ("127.0.0.1", 9002), ("127.0.0.1", 9003)}

    def test_update_should_replace_readers(self):
        self.table.update(self.new_table)
        assert self.table.readers == {("127.0.0.1", 9004), ("127.0.0.1", 9005)}

    def test_update_should_replace_writers(self):
        self.table.update(self.new_table)
        assert self.table.writers == {("127.0.0.1", 9006)}

    def test_update_should_replace_ttl(self):
        self.table.update(self.new_table)
        assert self.table.ttl == 300



# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import neo4j
import testkitbackend.fromtestkit as fromtestkit
import testkitbackend.totestkit as totestkit


def NewDriver(backend, data):
    authToken = data["authorizationToken"]["data"]
    auth = neo4j.Auth(
            authToken["scheme"], authToken["principal"],
            authToken["credentials"], realm=authToken["realm"])
    driver = neo4j.GraphDatabase.driver(data["uri"], auth=auth)
    key = backend.next_key()
    backend.drivers[key] = driver
    backend.send_response("Driver", {"id": key})


def DriverClose(backend, data):
    key = data["driverId"]
    driver = backend.drivers[key]
    driver.close()
    backend.send_response("Driver", {"id": key})


class SessionState:
    """ Keeps some extra state about the tracked session
    """

    def __init__(self, session):
        self.session = session


def NewSession(backend, data):
    driver = backend.drivers[data["driverId"]]
    access_mode = data["accessMode"]
    if access_mode == "r":
        access_mode = neo4j.READ_ACCESS
    elif access_mode == "w":
        access_mode = neo4j.WRITE_ACCESS
    else:
        raise Exception("Unknown access mode:" + access_mode)
    config = {
            "default_access_mode": access_mode,
            "bookmarks": data["bookmarks"],
            "database": data["database"],
    }
    # TODO: fetchSize, database
    session = driver.session(**config)
    key = backend.next_key()
    backend.sessions[key] = SessionState(session)
    backend.send_response("Session", {"id": key})


def SessionRun(backend, data):
    session = backend.sessions[data["sessionId"]].session
    cypher, params = fromtestkit.toCypherAndParams(data)
    # TODO: txMeta, timeout
    result = session.run(cypher, parameters=params)
    key = backend.next_key()
    backend.results[key] = result
    backend.send_response("Result", {"id": key})


def SessionClose(backend, data):
    key = data["sessionId"]
    session = backend.sessions[key].session
    session.close()
    del backend.sessions[key]
    backend.send_response("Session", {"id": key})


def SessionBeginTransaction(backend, data):
    key = data["sessionId"]
    session = backend.sessions[key].session
    metadata = data.get('txMeta', None)
    timeout = data.get('timeout', None)
    if timeout:
        timeout = int(timeout)
    tx = session.begin_transaction(metadata=metadata, timeout=timeout)
    key = backend.next_key()
    backend.transactions[key] = tx
    backend.send_response("Transaction", {"id": key})


def ResultNext(backend, data):
    result = backend.results[data["resultId"]]
    try:
        record = next(iter(result))
    except StopIteration:
        backend.send_response("NullRecord", {})
    backend.send_response("Record", totestkit.record(record))


def TransactionRun(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    cypher, params = fromtestkit.toCypherAndParams(data)
    result = tx.run(cypher, parameters=params)
    key = backend.next_key()
    backend.results[key] = result
    backend.send_response("Result", {"id": key})


def TransactionCommit(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    tx.commit()
    backend.send_response("Transaction", {"id": key})


def TransactionRollback(backend, data):
    key = data["txId"]
    tx = backend.transactions[key]
    tx.rollback()
    backend.send_response("Transaction", {"id": key})



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from unittest import TestCase
import pytest
from threading import (
    Thread,
    Event,
)
from neo4j import (
    Config,
    PoolConfig,
    WorkspaceConfig,
)
from neo4j.io import (
    Bolt,
    BoltPool,
    IOPool
)
from neo4j.exceptions import (
    ClientError,
    ServiceUnavailable,
)


class FakeSocket:
    def __init__(self, address):
        self.address = address

    def getpeername(self):
        return self.address

    def sendall(self, data):
        return

    def close(self):
        return


class QuickConnection:

    def __init__(self, socket):
        self.socket = socket
        self.address = socket.getpeername()

    def reset(self):
        pass

    def close(self):
        self.socket.close()

    def closed(self):
        return False

    def defunct(self):
        return False

    def timedout(self):
        return False


class FakeBoltPool(IOPool):

    def __init__(self, address, *, auth=None, **config):
        self.pool_config, self.workspace_config = Config.consume_chain(config, PoolConfig, WorkspaceConfig)
        if config:
            raise ValueError("Unexpected config keys: %s" % ", ".join(config.keys()))

        def opener(addr, timeout):
            return QuickConnection(FakeSocket(addr))

        super().__init__(opener, self.pool_config, self.workspace_config)
        self.address = address

    def acquire(self, access_mode=None, timeout=None, database=None):
        return self._acquire(self.address, timeout)


class BoltTestCase(TestCase):

    def test_open(self):
        with pytest.raises(ServiceUnavailable):
            connection = Bolt.open(("localhost", 9999), auth=("test", "test"))

    def test_open_timeout(self):
        with pytest.raises(ServiceUnavailable):
            connection = Bolt.open(("localhost", 9999), auth=("test", "test"), timeout=1)

    def test_ping(self):
        protocol_version = Bolt.ping(("localhost", 9999))
        assert protocol_version is None

    def test_ping_timeout(self):
        protocol_version = Bolt.ping(("localhost", 9999), timeout=1)
        assert protocol_version is None


class ConnectionPoolTestCase(TestCase):

    def setUp(self):
        self.pool = FakeBoltPool(("127.0.0.1", 7687))

    def tearDown(self):
        self.pool.close()

    def assert_pool_size(self, address, expected_active, expected_inactive, pool=None):
        if pool is None:
            pool = self.pool
        try:
            connections = pool.connections[address]
        except KeyError:
            self.assertEqual(0, expected_active)
            self.assertEqual(0, expected_inactive)
        else:
            self.assertEqual(expected_active, len([cx for cx in connections if cx.in_use]))
            self.assertEqual(expected_inactive, len([cx for cx in connections if not cx.in_use]))

    def test_can_acquire(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        assert connection.address == address
        self.assert_pool_size(address, 1, 0)

    def test_can_acquire_twice(self):
        address = ("127.0.0.1", 7687)
        connection_1 = self.pool._acquire(address, timeout=3)
        connection_2 = self.pool._acquire(address, timeout=3)
        assert connection_1.address == address
        assert connection_2.address == address
        assert connection_1 is not connection_2
        self.assert_pool_size(address, 2, 0)

    def test_can_acquire_two_addresses(self):
        address_1 = ("127.0.0.1", 7687)
        address_2 = ("127.0.0.1", 7474)
        connection_1 = self.pool._acquire(address_1, timeout=3)
        connection_2 = self.pool._acquire(address_2, timeout=3)
        assert connection_1.address == address_1
        assert connection_2.address == address_2
        self.assert_pool_size(address_1, 1, 0)
        self.assert_pool_size(address_2, 1, 0)

    def test_can_acquire_and_release(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        self.assert_pool_size(address, 1, 0)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)

    def test_releasing_twice(self):
        address = ("127.0.0.1", 7687)
        connection = self.pool._acquire(address, timeout=3)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)
        self.pool.release(connection)
        self.assert_pool_size(address, 0, 1)

    def test_in_use_count(self):
        address = ("127.0.0.1", 7687)
        self.assertEqual(self.pool.in_use_connection_count(address), 0)
        connection = self.pool._acquire(address, timeout=3)
        self.assertEqual(self.pool.in_use_connection_count(address), 1)
        self.pool.release(connection)
        self.assertEqual(self.pool.in_use_connection_count(address), 0)

    def test_max_conn_pool_size(self):
        with FakeBoltPool((), max_connection_pool_size=1) as pool:
            address = ("127.0.0.1", 7687)
            pool._acquire(address, timeout=0)
            self.assertEqual(pool.in_use_connection_count(address), 1)
            with self.assertRaises(ClientError):
                pool._acquire(address, timeout=0)
            self.assertEqual(pool.in_use_connection_count(address), 1)

    def test_multithread(self):
        with FakeBoltPool((), max_connection_pool_size=5) as pool:
            address = ("127.0.0.1", 7687)
            releasing_event = Event()

            # We start 10 threads to compete connections from pool with size of 5
            threads = []
            for i in range(10):
                t = Thread(target=acquire_release_conn, args=(pool, address, releasing_event))
                t.start()
                threads.append(t)

            # The pool size should be 5, all are in-use
            self.assert_pool_size(address, 5, 0, pool)
            # Now we allow thread to release connections they obtained from pool
            releasing_event.set()

            # wait for all threads to release connections back to pool
            for t in threads:
                t.join()
            # The pool size is still 5, but all are free
            self.assert_pool_size(address, 0, 5, pool)


def acquire_release_conn(pool, address, releasing_event):
    conn = pool._acquire(address, timeout=3)
    releasing_event.wait()
    pool.release(conn)



# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

def toCypherAndParams(data):
    cypher = data["cypher"]
    params = data["params"]
    # Optional
    if params is None:
        return cypher, None
    # Transform the params to Python native
    for p in params:
        params[p] = toParam(params[p])
    return cypher, params


def toParam(m):
    """ Converts testkit parameter format to driver (python) parameter
    """
    data = m["data"]
    name = m["name"]
    if name == "CypherString":
        return str(data["value"])
    if name == "CypherBool":
        return bool(data["value"])
    if name == "CypherInt":
        return int(data["value"])
    if name == "CypherFloat":
        return float(data["value"])
    if name == "CypherString":
        return str(data["value"])
    if name == "CypherNull":
        return None
    if name == "CypherList":
        ls = []
        for x in data["value"]:
            ls.append(toParam(x))
        return ls
    if name == "CypherMap":
        mp = {}
        for k, v in data["value"].items():
            mp[k] = toParam(v)
        return mp
    raise Exception("Unknown param type " + name)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from socket import (
    getaddrinfo,
    getservbyname,
    SOCK_STREAM,
    AF_INET,
    AF_INET6,
)


class Address(tuple):

    @classmethod
    def from_socket(cls, socket):
        address = socket.getpeername()
        return cls(address)

    @classmethod
    def parse(cls, s, default_host=None, default_port=None):
        if not isinstance(s, str):
            raise TypeError("Address.parse requires a string argument")
        if s.startswith("["):
            # IPv6
            host, _, port = s[1:].rpartition("]")
            port = port.lstrip(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0, 0, 0))
        else:
            # IPv4
            host, _, port = s.partition(":")
            try:
                port = int(port)
            except (TypeError, ValueError):
                pass
            return cls((host or default_host or "localhost",
                        port or default_port or 0))

    @classmethod
    def parse_list(cls, *s, default_host=None, default_port=None):
        """ Parse a string containing one or more socket addresses, each
        separated by whitespace.
        """
        if not all(isinstance(s0, str) for s0 in s):
            raise TypeError("Address.parse_list requires a string argument")
        return [Address.parse(a, default_host, default_port)
                for a in " ".join(s).split()]

    def __new__(cls, iterable):
        if isinstance(iterable, cls):
            return iterable
        n_parts = len(iterable)
        inst = tuple.__new__(cls, iterable)
        if n_parts == 2:
            inst.__class__ = IPv4Address
        elif n_parts == 4:
            inst.__class__ = IPv6Address
        else:
            raise ValueError("Addresses must consist of either "
                             "two parts (IPv4) or four parts (IPv6)")
        return inst

    #: Address family (AF_INET or AF_INET6)
    family = None

    def __repr__(self):
        return "{}({!r})".format(self.__class__.__name__, tuple(self))

    @property
    def host(self):
        return self[0]

    @property
    def port(self):
        return self[1]

    @classmethod
    def _dns_resolve(cls, address, family=0):
        """ Regular DNS resolver. Takes an address object and optional
        address family for filtering.

        :param address:
        :param family:
        :return:
        """
        try:
            info = getaddrinfo(address.host, address.port, family, SOCK_STREAM)
        except OSError:
            raise ValueError("Cannot resolve address {}".format(address))
        else:
            resolved = []
            for fam, _, _, _, addr in info:
                if fam == AF_INET6 and addr[3] != 0:
                    # skip any IPv6 addresses with a non-zero scope id
                    # as these appear to cause problems on some platforms
                    continue
                if addr not in resolved:
                    resolved.append(Address(addr))
            return resolved

    def resolve(self, family=0, resolver=None):
        """ Carry out domain name resolution on this Address object.

        If a resolver function is supplied, and is callable, this is
        called first, with this object as its argument. This may yield
        multiple output addresses, which are chained into a subsequent
        regular DNS resolution call. If no resolver function is passed,
        the DNS resolution is carried out on the original Address
        object.

        This function returns a list of resolved Address objects.

        :param family: optional address family to filter resolved
                       addresses by (e.g. AF_INET6)
        :param resolver: optional customer resolver function to be
                         called before regular DNS resolution
        """
        resolved = []
        if resolver:
            for address in map(Address, resolver(self)):
                resolved.extend(self._dns_resolve(address, family))
        else:
            resolved.extend(self._dns_resolve(self, family))
        return resolved

    @property
    def port_number(self):
        try:
            return getservbyname(self[1])
        except (OSError, TypeError):
            # OSError: service/proto not found
            # TypeError: getservbyname() argument 1 must be str, not X
            try:
                return int(self[1])
            except (TypeError, ValueError) as e:
                raise type(e)("Unknown port value %r" % self[1])


class IPv4Address(Address):

    family = AF_INET

    def __str__(self):
        return "{}:{}".format(*self)


class IPv6Address(Address):

    family = AF_INET6

    def __str__(self):
        return "[{}]:{}".format(*self)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# Copyright (c) 2002-2020 "Neo4j,"
# Neo4j Sweden AB [http://neo4j.com]
#
# This file is part of Neo4j.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from neo4j import (
    GraphDatabase,
)
from neo4j.api import READ_ACCESS

from tests.stub.conftest import StubCluster

# python -m pytest tests/stub/test_bookmarking.py -s -v


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_should_be_no_bookmark_in_new_session(driver_info, test_script):
    # python -m pytest tests/stub/test_bookmarking.py -s -v -k test_should_be_no_bookmark_in_new_session
    with StubCluster(test_script):
        uri = "neo4j://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(fetch_size=-1) as session:
                assert session.last_bookmark() is None


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_should_be_able_to_set_bookmark(driver_info, test_script):
    # python -m pytest tests/stub/test_bookmarking.py -s -v -k test_should_be_able_to_set_bookmark
    with StubCluster(test_script):
        uri = "neo4j://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(bookmarks=["X"], fetch_size=-1) as session:
                assert session._bookmarks == ("X",)


@pytest.mark.parametrize(
    "test_script",
    [
        "v3/router.script",
        "v4x0/router.script",
    ]
)
def test_should_be_able_to_set_multiple_bookmarks(driver_info, test_script):
    # python -m pytest tests/stub/test_bookmarking.py -s -v -k test_should_be_able_to_set_multiple_bookmarks
    with StubCluster(test_script):
        uri = "neo4j://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(bookmarks=[":1", ":2"], fetch_size=-1) as session:
                assert session._bookmarks == (":1", ":2")


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/bookmark_chain.script"),
        ("v4x0/router.script", "v4x0/tx_bookmark_chain.script"),
    ]
)
def test_should_automatically_chain_bookmarks(driver_info, test_scripts):
    # python -m pytest tests/stub/test_bookmarking.py -s -v -k test_should_automatically_chain_bookmarks
    with StubCluster(*test_scripts):
        uri = "neo4j://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, bookmarks=["bookmark:0", "bookmark:1"], fetch_size=-1) as session:
                with session.begin_transaction():
                    pass
                assert session.last_bookmark() == "bookmark:2"
                with session.begin_transaction():
                    pass
                assert session.last_bookmark() == "bookmark:3"


@pytest.mark.parametrize(
    "test_scripts",
    [
        ("v3/router.script", "v3/bookmark_chain_with_autocommit.script"),
        ("v4x0/router.script", "v4x0/tx_bookmark_chain_with_autocommit.script"),
    ]
)
def test_autocommit_transaction_included_in_chain(driver_info, test_scripts):
    # python -m pytest tests/stub/test_bookmarking.py -s -v -k test_autocommit_transaction_included_in_chain
    with StubCluster(*test_scripts):
        uri = "neo4j://localhost:9001"
        with GraphDatabase.driver(uri, auth=driver_info["auth_token"]) as driver:
            with driver.session(default_access_mode=READ_ACCESS, bookmarks=["bookmark:1"], fetch_size=-1) as session:
                with session.begin_transaction():
                    pass
                assert session.last_bookmark() == "bookmark:2"
                session.run("RETURN 1").consume()
                assert session.last_bookmark() == "bookmark:3"
                with session.begin_transaction():
                    pass
                assert session.last_bookmark() == "bookmark:4"
